(*  Title:      HOL/Tools/BNF/bnf_nu_fp_ind_sugar.ML
    Author:     Jasmin Blanchette, Inria, LORIA, MPII
    Copyright   2016

Sugared (co)induction over nonuniform (co)datatypes.
*)

signature BNF_NU_FP_IND_SUGAR =
sig
  val gen_nonuninform_induct: string list -> ((binding * Token.src list) * term) list ->
    local_theory ->
    thm list * (term * 'a list) list list
    * (thm list list -> local_theory -> local_theory) * local_theory
  val nonuninform_induct_cmd: string list -> Specification.multi_specs_cmd -> local_theory ->
    Proof.state

  val gen_nonuninform_coinduct: term list -> ((binding * Token.src list) * term) list ->
    local_theory ->
    thm list * (term * 'a list) list list
    * (thm list list -> Proof.context -> local_theory) * local_theory
  val nonuninform_coinduct_cmd: string list -> Specification.multi_specs_cmd -> local_theory ->
    Proof.state
end;

structure BNF_NU_FP_Ind_Sugar : BNF_NU_FP_IND_SUGAR =
struct

open BNF_Util
open BNF_Def
open BNF_FP_Rec_Sugar_Util
open BNF_FP_Def_Sugar
open BNF_NU_FP_Sugar_Tactics
open BNF_NU_FP_Def_Sugar
open BNF_NU_FP_Ind

val P_N = "P";
val Q_N = "Q";

fun gen_nonuninform_induct on_var_names specs lthy =
  let
    val nn = length on_var_names;

    val _ = nn > 0 orelse error "No induction variable";
    val _ = length on_var_names = length specs orelse
      error "Different number of variables and properties";

    val result_thm_names = map fst specs;
    val props = map snd specs;

    fun extract_P_rhss on_var_name prop =
      let val atom_prop = Object_Logic.atomize_term lthy prop in
        if fastype_of atom_prop = HOLogic.boolT then
          let
            fun maybe_add_free (t as Free (s, _)) = s = on_var_name ? insert (op aconv) t
              | maybe_add_free _ = I;

            val frees = fold_aterms maybe_add_free atom_prop [];
          in
            (case frees of
              [] => error ("Free variable " ^ quote on_var_name ^ " does not occur in property " ^
                Syntax.string_of_term lthy prop)
            | [free as Free (_, T)] =>
              let
                fun not_nu_datatype () =
                  error ("Type of variable " ^ quote (Syntax.string_of_term lthy free) ^
                    " not a nonuniform datatype");
              in
                (case T of
                  Type (T_name, _) =>
                  (case nu_fp_sugar_of lthy T_name of
                    SOME {fp_sugar = {fp = Least_FP, ...}, ...} => lambda free atom_prop
                  | _ => not_nu_datatype ())
                | _ => not_nu_datatype ())
              end
            | _ => error ("Ambiguous free variable name " ^ quote on_var_name ^ " in property " ^
                Syntax.string_of_term lthy prop))
          end
        else
          error ("Cannot atomize property " ^ quote (Syntax.string_of_term lthy prop))
      end;

    val P_rhss = map2 extract_P_rhss on_var_names props;

    val fpTs0 = map (domain_type o fastype_of) P_rhss;
    val fpT_names = map (fst o dest_Type) fpTs0;

    val _ = (case duplicates (op =) fpT_names of [] => ()
      | fpT_name :: _ => error ("Several induction variables for type " ^ quote fpT_name));

    val nu_fp_sugars as {T = Type (_, any_fpT_args0), fp_sugar = {fp_nesting_bnfs, ...},
        nu_fp_res as {fp_res = {Ts = res_fpTs, bnfs = any_fp_bnf :: _, ctors = ctors0, ...}, ...},
        ...} :: _ =
      map (the o nu_fp_sugar_of lthy) fpT_names;
    val fp_sugars = map #fp_sugar nu_fp_sugars;

    val _ =
      forall (curry (op = o apply2 (#Ts o #fp_res)) nu_fp_res o #nu_fp_res) (tl nu_fp_sugars) orelse
      error ("Types " ^ commas (map quote fpT_names) ^
        " are not part of the same mutually recursive definition");
    val _ = nn = length res_fpTs orelse
      error ("Expected induction on " ^ string_of_int (length res_fpTs) ^ " variables");

    val num_As = length any_fpT_args0;
    val liveness = liveness_of_fp_bnf num_As any_fp_bnf;
    val live = live_of_bnf any_fp_bnf;

    val ((As0, Bs0), _) = lthy
      |> mk_TFrees num_As
      ||>> mk_TFrees num_As;

    val Xss = map #Xs nu_fp_sugars;
    val Xs = flat Xss;

    val As = map2 (resort_tfree_or_tvar o Type.sort_of_atyp) any_fpT_args0 As0;
    val Bs = @{map 3} (fn alive => fn A as TFree (_, S) => fn B =>
        if alive then resort_tfree_or_tvar S B else A)
      liveness As Bs0;
    val (live_AsBs, dead_As) = List.partition (op <>) (As ~~ Bs) ||> map fst;

    fun stutter xs = flat (map2 (map o K) xs Xss);

    val (Rs, names_lthy) = lthy
      |> mk_Frees "R" (map (uncurry mk_pred2T) live_AsBs);

    val P_bs = map2 (fn (thm_name, _) => fn fpT_name =>
        Binding.qualify_name false thm_name (P_N ^ "_" ^ Long_Name.base_name fpT_name))
      result_thm_names fpT_names;

    val ((raw_P_consts, raw_P_defs), (lthy, lthy_old)) = lthy
      |> Local_Theory.open_target |> snd
      |> apfst split_list o @{fold_map 2} (fn b => fn rhs =>
          Local_Theory.define ((b, NoSyn), ((Thm.def_binding b, []), rhs))
          #>> apsnd snd)
        P_bs P_rhss
      ||> `Local_Theory.close_target;

    val phi = Proof_Context.export_morphism lthy_old lthy;

    val P_defs = map (Morphism.thm phi) raw_P_defs;

    val P_consts = map (Morphism.term phi) raw_P_consts;
    val PA_consts = map (mk_dtor As) P_consts;
    val PB_consts = map (mk_dtor Bs) P_consts;

    fun mk_bi_unique R =
      HOLogic.mk_Trueprop (Const (@{const_name bi_unique}, mk_pred1T (fastype_of R)) $ R);
    fun mk_left_total R =
      HOLogic.mk_Trueprop (Const (@{const_name left_total}, mk_pred1T (fastype_of R)) $ R);
    fun weaken_parametricity_goal (Trueprop $ (rel_fun $ R $ eq $ PA $ PB)) =
      Trueprop $ (rel_fun $ R $ @{term "op \<longleftarrow>"} $ PA $ PB);

    fun mk_bi_unique_left_total_parametricity_goal Rs PA PB =
      Logic.list_implies (map mk_bi_unique Rs @ map mk_left_total Rs,
        weaken_parametricity_goal (mk_parametricity_goal lthy Rs PA PB));

    val parametricity_goals =
      map2 (rpair [] oo mk_bi_unique_left_total_parametricity_goal Rs) PA_consts PB_consts;

    val live_arg_absT_infosss = map #live_arg_absT_infoss nu_fp_sugars;
    val pre_bnfs = map #pre_bnf fp_sugars;
    val fp_bnfs = map #fp_bnf fp_sugars;

    val ctrss0 = map (#ctrs o #ctr_sugar o #fp_ctr_sugar) fp_sugars;
    val ctr0_Tsss = map (map (binder_types o fastype_of)) ctrss0;
    val ctrXs0_Tsss = map (#ctrXs_Tss o #fp_ctr_sugar) fp_sugars;

    val nn' = length Xs;
    val ns = map length ctr0_Tsss;
    val mss = map (map length) ctr0_Tsss;

    val live_arg_type_definitionsss = map (map (map #type_definition)) live_arg_absT_infosss;
    val live_arg_abs_injectsss = map (map (map #abs_inject)) live_arg_absT_infosss;
    val live_arg_rep_injectsss =
      map (map (map (fn thm => thm RS @{thm type_definition.Rep_inject})))
        live_arg_type_definitionsss;
    val live_arg_abs_inversesss = map (map (map #abs_inverse)) live_arg_absT_infosss;
    val live_arg_rep_inversesss =
      map (map (map (fn thm => thm RS @{thm type_definition.Rep_inverse})))
        live_arg_type_definitionsss;
    val live_arg_rep_inverses = flat (flat live_arg_rep_inversesss);
    val pre_abs_inverses = map (#abs_inverse o #absT_info) fp_sugars;
    val pre_type_definitions = map (#type_definition o #absT_info) fp_sugars;
    val ctr_defss = map (#ctr_defs o #fp_ctr_sugar) fp_sugars;
    val pre_map_comps = map map_comp_of_bnf pre_bnfs;
    val pre_map_idents = map map_ident_of_bnf pre_bnfs;
    val pre_pred_sets = map pred_set_of_bnf pre_bnfs;
    val pre_set_defss = map set_defs_of_bnf pre_bnfs;
    val pre_set_maps = maps set_map_of_bnf pre_bnfs;
    val fp_map_comps = map map_comp_of_bnf fp_bnfs;
    val fp_map_idents = map map_ident_of_bnf fp_bnfs;
    val fp_map_congs = map map_cong_of_bnf fp_bnfs;
    val fp_inj_maps = map inj_map_of_bnf fp_bnfs;
    val fp_nesting_set_maps = maps set_map_of_bnf fp_nesting_bnfs;

    val setss_pre = map mk_bnf_sets pre_bnfs;
    val setss_fp_nesting = map mk_bnf_sets fp_nesting_bnfs;

    val (P_ctr_goalss, kksss) =
      let
        val raw_goalss =
          @{map 4} (@{map 3} o mk_induct_raw_prem I names_lthy (map single Xs) setss_fp_nesting)
            P_consts ctrss0 ctr0_Tsss ctrXs0_Tsss;
        val goalss = map (map (finish_induct_prem lthy 0 (stutter P_consts))) raw_goalss;
        val kksss = map (map (map (fst o snd) o #2)) raw_goalss;
      in
        (map (map (rpair [])) goalss, kksss)
      end;

    val goalss = parametricity_goals :: P_ctr_goalss;

    fun mk_absrep_tuple {absT = Type (abs_Tname, _), abs = abs0, rep = rep0, ...} =
      let
        val absT = Type (abs_Tname, As);
        val abs = enforce_type lthy range_type absT abs0;
        val rep = enforce_type lthy domain_type absT rep0;
        val repT = range_type (fastype_of rep);
      in (absT, (repT, (abs, rep))) end;

    val live_arg_absrep_tuples = maps (maps (map mk_absrep_tuple)) live_arg_absT_infosss;
    val live_arg_absTs = map fst live_arg_absrep_tuples;
    val live_arg_absrepTs = map (apsnd fst) live_arg_absrep_tuples;
    val live_arg_absTabs = map (apsnd (fst o snd)) live_arg_absrep_tuples;
    val live_arg_absTrep = map (apsnd (snd o snd)) live_arg_absrep_tuples;

    fun X_ify_preT (Type (s, Ts)) = Type (s, take (length Ts - length Xs) Ts @ Xs);

    val ctors = map (mk_ctor As) ctors0;
    val ctor_Ts = map (domain_type o fastype_of) ctors;
    val ctor0_Ts = map (typ_subst_nonatomic live_arg_absrepTs) ctor_Ts;
    val ctor0ctor_Ts = ctor0_Ts ~~ ctor_Ts;
    val ctorXs_Ts = map X_ify_preT ctor_Ts;

    fun after_qed (parametricity_thms :: P_ctr_thmss) lthy =
      let
        val P_ctor0_thms =
          let
            fun absify_x x =
              let
                val T0 = fastype_of x;
                val T = the (AList.lookup (op =) ctor0ctor_Ts T0);
              in
                build_map lthy [] live_arg_absTs (the o AList.lookup (op =) live_arg_absTabs o snd)
                  (T0, T)
                $ x
              end;

            val raw_goals = @{map 4} (mk_induct_raw_prem absify_x names_lthy Xss setss_pre)
              P_consts ctors (map single ctor0_Ts) (map single ctorXs_Ts);
            val goals = map (finish_induct_prem lthy 0 P_consts) raw_goals;

            fun prove n ms kkss pre_type_definition pre_set_defs ctr_defs P_ctr_thms goal =
              Goal.prove_sorry lthy [] [] goal (fn {context = ctxt, prems = _} =>
                mk_nonuniform_induct_P_ctor0_tac ctxt nn' n ms kkss ctr_defs pre_type_definition
                  P_ctr_thms pre_abs_inverses fp_nesting_set_maps pre_set_defs)
              |> Thm.close_derivation;
          in
            @{map 8} prove ns mss kksss pre_type_definitions pre_set_defss ctr_defss P_ctr_thmss
              goals
          end;

        fun mk_P_ctor_thms P_mono_thms =
          let
            val raw_goals = @{map 4} (mk_induct_raw_prem I names_lthy Xss setss_pre)
              P_consts ctors (map single ctor_Ts) (map single ctorXs_Ts);
            val goals = map (finish_induct_prem lthy 0 P_consts) raw_goals;

            fun prove T0 T P_ctor0_thm goal =
              let
                val x = Var ((Name.uu, 0), T);
                val map_rep_x = build_map lthy live_arg_absTs []
                    (the o AList.lookup (op =) live_arg_absTrep o fst) (T, T0)
                  $ x
                  |> (snd o Type.varify_global []);

                val P_ctor0_thm' = P_ctor0_thm
                  |> infer_instantiate' lthy [SOME (Thm.cterm_of lthy map_rep_x)]
                  |> unfold_thms lthy (pre_set_maps @ pre_map_comps @ fp_map_comps @
                    pre_map_idents @ fp_map_idents @ live_arg_rep_inverses @ @{thms o_def id_def});
              in
                Goal.prove_sorry lthy [] [] goal (fn {context = ctxt, prems = _} =>
                  mk_nonuniform_induct_P_ctor_tac ctxt fp_map_comps fp_map_idents fp_inj_maps
                    (stutter P_mono_thms) (flat live_arg_abs_injectsss)
                    (flat live_arg_rep_injectsss) (flat live_arg_abs_inversesss) P_ctor0_thm')
                |> Thm.close_derivation
              end;
          in
            @{map 4} prove ctor0_Ts ctor_Ts P_ctor0_thms goals
          end;

        fun mk_ind0_thms P_mono_thms =
          let
            val P_ctor_thms = mk_P_ctor_thms P_mono_thms;

            fun mk_thm ctor_T pre_bnf ctor PA_const pre_pred_set P_ctor_thm =
              let
                val ctor_predT = mk_pred1T ctor_T;

                val pre_pred =
                  enforce_type lthy (domain_type o body_fun_type) ctor_T (pred_of_bnf pre_bnf);
                val (true_pred_domTs, P_domTs) =
                  chop live (map domain_type (binder_fun_types (fastype_of pre_pred)));
                val true_preds = map (fn T => Abs (Name.uu, T, @{term True})) true_pred_domTs;
                val P_consts' = map2 (enforce_type lthy domain_type) P_domTs (stutter P_consts);
                val leq = Const (@{const_name less_eq}, mk_pred2T ctor_predT ctor_predT);

                val lhs = list_comb (pre_pred, true_preds @ P_consts');
                val rhs = HOLogic.mk_comp (PA_const, ctor);
                val goal = HOLogic.mk_Trueprop (leq $ lhs $ rhs);
              in
                Goal.prove_sorry lthy [] [] goal (fn {context = ctxt, prems = _} =>
                  mk_nonuniform_ind0_tac ctxt pre_pred_set P_ctor_thm)
                |> Thm.close_derivation
              end
          in
            @{map 6} mk_thm ctor_Ts pre_bnfs ctors PA_consts pre_pred_sets P_ctor_thms
          end;

        val ({co_ind_thms = P_thms}, lthy) =
          construct_nu_co_ind Least_FP nu_fp_res dead_As P_consts parametricity_thms mk_ind0_thms
            lthy;
        val result_thms = map (unfold_thms lthy P_defs) P_thms;

        val notes = map2 (fn thm => rpair [([thm], [])]) result_thms result_thm_names;

        val lthy = lthy |> Local_Theory.notes notes |> snd;
      in
        lthy
      end;
  in
    (P_defs, goalss, after_qed, lthy)
  end;

val nonuninform_induct_cmd = (fn (P_defs, goalss, after_qed, lthy) =>
  lthy
  |> Proof.theorem NONE after_qed goalss
  |> Proof.unfolding ([[(P_defs, [])]])
  |> Proof.refine_singleton (Method.primitive_text (K I))) ooo
    (fn on_var_names => fn raw_specs => fn lthy =>
       let
         val specs = snd (fst (Specification.read_multi_specs [] raw_specs lthy))
           |> map (apsnd drop_all);
       in
         gen_nonuninform_induct on_var_names specs lthy
       end);

val _ = Outer_Syntax.local_theory_to_proof @{command_keyword nonuniform_induct}
  "perform induction on nonuniform datatypes"
  (Scan.repeat1 Parse.short_ident --| Parse.$$$ "in"
   -- Parse_Spec.multi_specs >> uncurry nonuninform_induct_cmd);

fun gen_nonuninform_coinduct bisims specs lthy =
  let
    val nn = length bisims;

    val _ = nn > 0 orelse error "No coinduction relation";
    val _ = length bisims = length specs orelse
      error "Different number of relations and equations";

    val result_thm_names = map fst specs;
    val eqs = map snd specs;

    fun dest_Trueprop_eq (@{const Trueprop} $ (Const (@{const_name HOL.eq}, _) $ lhs $ rhs)) =
        (lhs, rhs)
      | dest_Trueprop_eq prop =
        error ("Property " ^ quote (Syntax.string_of_term lthy prop) ^ " not a HOL equation");

    fun is_possibly_bisim_T
        (Type (@{type_name fun}, [T1, Type (@{type_name fun}, [T2, @{typ bool}])])) = (T1 = T2)
      | is_possibly_bisim_T _ = false;

    fun extract_Q_rhss bisim lhs rhs =
      let val T = fastype_of lhs in
        if is_possibly_bisim_T (fastype_of bisim) then
          let
            fun not_nu_codatatype () =
              error ("Type of terms " ^ quote (Syntax.string_of_term lthy lhs) ^ ", " ^
                quote (Syntax.string_of_term lthy rhs) ^ " not a nonuniform codatatype");
          in
            (case T of
              Type (T_name, _) =>
              (case nu_fp_sugar_of lthy T_name of
                SOME {fp_sugar = {fp = Greatest_FP, ...}, ...} => bisim
              | _ => not_nu_codatatype ())
            | _ => not_nu_codatatype ())
          end
        else
          error ("Term " ^ quote (Syntax.string_of_term lthy bisim) ^ " must have type " ^
            quote (Syntax.string_of_typ lthy (mk_pred2T T T)))
      end;

    val (lhss, rhss) = split_list (map dest_Trueprop_eq eqs);
    val Q_rhss = @{map 3} extract_Q_rhss bisims lhss rhss;

    val fpTs0 = map (domain_type o fastype_of) Q_rhss;
    val fpT_names = map (fst o dest_Type) fpTs0;
    val fp_b_names = map base_name_of_typ fpTs0;

    val _ = (case duplicates (op =) fpT_names of [] => ()
      | fpT_name :: _ => error ("Several coinduction relations for type " ^ quote fpT_name));

    val nu_fp_sugars as {T = Type (_, any_fpT_args0), fp_sugar = {live_nesting_bnfs, ...},
        nu_fp_res as {fp_res = {Ts = res_fpTs, bnfs = any_fp_bnf :: _, dtors = dtors0, dtor_ctors,
          ...}, ...}, ...} :: _ =
      map (the o nu_fp_sugar_of lthy) fpT_names;
    val fp_sugars = map #fp_sugar nu_fp_sugars;
    val ctr_sugars = map (#ctr_sugar o #fp_ctr_sugar) fp_sugars;

    val _ =
      forall (curry (op = o apply2 (#Ts o #fp_res)) nu_fp_res o #nu_fp_res) (tl nu_fp_sugars) orelse
      error ("Types " ^ commas (map quote fpT_names) ^
        " are not part of the same mutually corecursive definition");
    val _ = nn = length res_fpTs orelse
      error ("Expected coinduction with " ^ string_of_int (length res_fpTs) ^ " relations");

    val num_As = length any_fpT_args0;
    val liveness = liveness_of_fp_bnf num_As any_fp_bnf;
    val live = live_of_bnf any_fp_bnf;

    val ((As0, Bs0), _) = lthy
      |> mk_TFrees num_As
      ||>> mk_TFrees num_As;

    val As = map2 (resort_tfree_or_tvar o Type.sort_of_atyp) any_fpT_args0 As0;
    val Bs = @{map 3} (fn alive => fn A as TFree (_, S) => fn B =>
        if alive then resort_tfree_or_tvar S B else A)
      liveness As Bs0;
    val (live_AsBs, dead_As) = List.partition (op <>) (As ~~ Bs) ||> map fst;

    val substAT = Term.typ_subst_atomic (any_fpT_args0 ~~ As);
    val substA = Term.map_types substAT;

    val Xss = map #Xs nu_fp_sugars;
    val Xs = flat Xss;
    val real_Tss = map (map substAT o #real_Ts) nu_fp_sugars;
    val real_Ts = flat real_Tss;

    val (((Rs, us'), vs'), _) = lthy
      |> mk_Frees "R" (map (uncurry mk_pred2T) live_AsBs)
      ||>> Variable.variant_fixes fp_b_names
      ||>> Variable.variant_fixes (map (suffix "'") fp_b_names);

    val Q_bs = map2 (fn (thm_name, _) => fn fpT_name =>
        Binding.qualify_name false thm_name (Q_N ^ "_" ^ Long_Name.base_name fpT_name))
      result_thm_names fpT_names;

    val ((raw_Q_consts, raw_Q_defs), (lthy, lthy_old)) = lthy
      |> Local_Theory.open_target |> snd
      |> apfst split_list o @{fold_map 2} (fn b => fn rhs =>
          Local_Theory.define ((b, NoSyn), ((Thm.def_binding b, []), rhs))
          #>> apsnd snd)
        Q_bs Q_rhss
      ||> `Local_Theory.close_target;

    val phi = Proof_Context.export_morphism lthy_old lthy;

    val Q_defs = map (Morphism.thm phi) raw_Q_defs;

    val Q_consts = map (Morphism.term phi) raw_Q_consts;
    val QA_consts = map (mk_dtor As) Q_consts;
    val QB_consts = map (mk_dtor Bs) Q_consts;

    fun mk_bi_unique R =
      HOLogic.mk_Trueprop (Const (@{const_name bi_unique}, mk_pred1T (fastype_of R)) $ R);
    fun mk_left_total R =
      HOLogic.mk_Trueprop (Const (@{const_name left_total}, mk_pred1T (fastype_of R)) $ R);
    fun weaken_parametricity_goal (Trueprop $ (rel_fun $ R $ (rel_fun' $ S $ eq) $ PA $ PB)) =
      Trueprop $ (rel_fun $ R $ (rel_fun' $ S $ @{term "op \<longrightarrow>"}) $ PA $ PB);
    fun mk_bi_unique_left_total_parametricity_goal Rs QA QB =
      Logic.list_implies (map mk_bi_unique Rs @ map mk_left_total Rs,
        weaken_parametricity_goal (mk_parametricity_goal lthy Rs QA QB));

    val parametricity_goals =
      map2 (rpair [] oo mk_bi_unique_left_total_parametricity_goal Rs) QA_consts QB_consts;

    val live_arg_absT_infosss = map #live_arg_absT_infoss nu_fp_sugars;
    val pre_bnfs = map #pre_bnf fp_sugars;
    val fp_bnfs = map #fp_bnf fp_sugars;

    val ctrss0 = map #ctrs ctr_sugars;
    val ctr0_Tsss = map (map (binder_types o fastype_of)) ctrss0;
    val ctrXs_Tsss = map (map (map substAT) o #ctrXs_Tss o #fp_ctr_sugar) fp_sugars;
    val discss = map (map substA o #discs) ctr_sugars;
    val selsss = map (map (map substA) o #selss) ctr_sugars;

    val ns = map length ctr0_Tsss;

    val live_arg_type_definitionsss = map (map (map #type_definition)) live_arg_absT_infosss;
    val live_arg_rep_inversesss =
      map (map (map (fn thm => thm RS @{thm type_definition.Rep_inverse})))
        live_arg_type_definitionsss;
    val live_arg_abs_injectsss =
      map (map (map (fn thm =>
          thm RS @{thm type_definition.Abs_inject[OF _ UNIV_I UNIV_I]} RS iffD1)))
        live_arg_type_definitionsss;
    val live_arg_rep_inverses = flat (flat live_arg_rep_inversesss);
    val live_arg_abs_injects = flat (flat live_arg_abs_injectsss);
    val live_arg_abs_inversesss = map (map (map #abs_inverse)) live_arg_absT_infosss;
    val live_arg_abs_inverses = flat (flat live_arg_abs_inversesss);
    val pre_abs_inverses = map (#abs_inverse o #absT_info) fp_sugars;
    val ctr_defss = map (#ctr_defs o #fp_ctr_sugar) fp_sugars;
    val exhausts = map #exhaust ctr_sugars;
    val disc_thmsss = map #disc_thmss ctr_sugars;
    val sel_thmsss = map #sel_thmss ctr_sugars;
    val pre_rel_defs = map rel_def_of_bnf pre_bnfs;
    val pre_rel_maps = maps rel_map_of_bnf pre_bnfs;
    val pre_rel_mono_strongs = map rel_mono_strong_of_bnf pre_bnfs;
    val pre_map_comps = map map_comp_of_bnf pre_bnfs;
    val pre_map_idents = map map_ident_of_bnf pre_bnfs;
    val fp_map_comps = map map_comp_of_bnf fp_bnfs;
    val fp_map_idents = map map_ident_of_bnf fp_bnfs;
    val fp_map_congs = map map_cong_of_bnf fp_bnfs;
    val live_nesting_rel_eqs = map rel_eq_of_bnf live_nesting_bnfs;

    val dtors = map (mk_dtor As) dtors0;
    val fpTs = map (domain_type o fastype_of) dtors;
    val ctor_Ts = map (range_type o fastype_of) dtors;

    val us = map2 (curry Free) us' fpTs;
    val udiscss = map2 (map o rapp) us discss;
    val uselsss = map2 (map o map o rapp) us selsss;
    val dtor_udiscs = map (fn u => HOLogic.mk_eq (u, u)) us;
    val udtors = map2 rapp us dtors;

    val vs = map2 (curry Free) vs' fpTs;
    val vdiscss = map2 (map o rapp) vs discss;
    val vselsss = map2 (map o map o rapp) vs selsss;
    val dtor_vdiscs = map (fn v => HOLogic.mk_eq (v, v)) vs;
    val vdtors = map2 rapp vs dtors;

    val uvQs = @{map 3} (fn QA => fn u => fn v => QA $ u $ v) QA_consts us vs;

    val Q_discsel_goals =
      @{map 9} (mk_coinduct_prem lthy Xss real_Tss Q_consts) uvQs us vs ns udiscss uselsss vdiscss
        vselsss ctrXs_Tsss
      |> map (rpair []);

    fun mk_Q_lhsrhs_goal bisim Q lhs rhs =
      let val T = fastype_of lhs in
        (case try (enforce_type lthy domain_type T) Q of
          SOME Q' => HOLogic.mk_Trueprop (Q' $ lhs $ rhs)
        | NONE => error ("Term " ^ quote (Syntax.string_of_term lthy bisim) ^ " must have type " ^
            quote (Syntax.string_of_typ lthy (mk_pred2T T T))))
      end;

    val Q_lhsrhs_goals = @{map 4} (rpair [] oooo mk_Q_lhsrhs_goal) bisims Q_consts lhss rhss;

    val goalss = [parametricity_goals, Q_discsel_goals, Q_lhsrhs_goals];

    fun mk_absrep_tuple {absT = Type (abs_Tname, _), abs = abs0, rep = rep0, ...} =
      let
        val absT = Type (abs_Tname, As);
        val abs = enforce_type lthy range_type absT abs0;
        val rep = enforce_type lthy domain_type absT rep0;
        val repT = range_type (fastype_of rep);
      in (absT, (repT, (abs, rep))) end;

    val live_arg_absrep_tuples = maps (maps (map mk_absrep_tuple)) live_arg_absT_infosss;
    val live_arg_absTs = map fst live_arg_absrep_tuples;
    val live_arg_absTrep = map (apsnd (snd o snd)) live_arg_absrep_tuples;

    fun X_ify_preT (Type (s, Ts)) = Type (s, take (length Ts - length Xs) Ts @ Xs);

    val ctorXs_Ts = map X_ify_preT ctor_Ts;

    val substXs_abs_real =
      map_atyps (Type.devar (fold (Sign.typ_match (Proof_Context.theory_of lthy))
        (ctorXs_Ts ~~ ctor_Ts) Vartab.empty));
    val substXs_real = Term.typ_subst_atomic (Xs ~~ real_Ts);

    val abs_real_Tss = map (map substXs_abs_real) Xss;

    fun after_qed [parametricity_thms, Q_discsel_thms, Q_lhsrhs_thms] lthy =
      let
        val Q_dtor0_thms =
          let
            fun repify_pre pre =
              let
                val preT = fastype_of pre;
                val preU = substXs_real (X_ify_preT preT);
              in
                build_map lthy live_arg_absTs [] (the o AList.lookup (op =) live_arg_absTrep o fst)
                  (preT, preU)
                $ pre
              end;

            val goals = @{map 9} (mk_coinduct_prem lthy Xss real_Tss Q_consts)
              uvQs us vs (replicate nn 1) (map single dtor_udiscs)
              (map (single o single o repify_pre) udtors)
              (map single dtor_vdiscs) (map (single o single o repify_pre) vdtors)
              (map (single o single) ctorXs_Ts);

            fun prove kk n pre_rel_def pre_abs_inverse dtor_ctor exhaust ctr_defs disc_thmss
                sel_thmss goal =
              Goal.prove_sorry lthy [] [] goal (fn {context = ctxt, prems = _} =>
                mk_nonuniform_induct_Q_dtor0_tac ctxt Q_discsel_thms pre_map_comps pre_map_idents
                  fp_map_comps fp_map_idents live_arg_abs_inverses live_nesting_rel_eqs nn kk n
                  pre_rel_def pre_abs_inverse dtor_ctor exhaust ctr_defs disc_thmss sel_thmss)
              |> Thm.close_derivation;
          in
            @{map 10} prove (1 upto nn) ns pre_rel_defs pre_abs_inverses dtor_ctors exhausts
              ctr_defss disc_thmsss sel_thmsss goals
          end;

        fun mk_Q_dtor_thms Q_mono_thms =
          let
            val goals = @{map 9} (mk_coinduct_prem lthy Xss abs_real_Tss Q_consts)
              uvQs us vs (replicate nn 1) (map single dtor_udiscs) (map (single o single) udtors)
              (map single dtor_vdiscs) (map (single o single) vdtors)
              (map (single o single) ctorXs_Ts);

            fun prove Q_dtor0_thm goal =
              Goal.prove_sorry lthy [] [] goal (fn {context = ctxt, prems = _} =>
                mk_nonuniform_induct_Q_dtor_tac ctxt live (map length Xss) live_arg_abs_injects
                  live_arg_rep_inverses fp_map_comps fp_map_idents fp_map_congs pre_rel_maps
                  pre_rel_mono_strongs Q_dtor0_thm Q_mono_thms)
              |> Thm.close_derivation;
          in
            map2 prove Q_dtor0_thms goals
          end;

        fun mk_coind0_thms Q_mono_thms =
          let
            val objectify_Q_dtor =
              unfold_thms lthy @{thms objectify_coind_prem} o forall_intr_vars;

            fun mk_thm thm =
              (objectify_Q_dtor thm RS @{thm predicate2I_obj})
              |> unfold_thms lthy @{thms BNF_Def.vimage2p_def[symmetric]};
          in
            map mk_thm (mk_Q_dtor_thms Q_mono_thms)
          end;

        val ({co_ind_thms = Q_imp_thms}, lthy) =
          construct_nu_co_ind Greatest_FP nu_fp_res dead_As Q_consts parametricity_thms
            mk_coind0_thms lthy;

        fun derive_result Q_def Q_lhsrhs_thm Q_imp_thm =
          unfold_thms lthy [Q_def] (Q_lhsrhs_thm RS Q_imp_thm);

        val result_thms = @{map 3} derive_result Q_defs Q_lhsrhs_thms Q_imp_thms;

        val notes = map2 (fn thm => rpair [([thm], [])]) result_thms result_thm_names;

        val lthy = lthy |> Local_Theory.notes notes |> snd;
      in
        lthy
      end;
  in
    (Q_defs, goalss, after_qed, lthy)
  end;

val nonuninform_coinduct_cmd = (fn (Q_defs, goalss, after_qed, lthy) =>
  lthy
  |> Proof.theorem NONE after_qed goalss
  |> Proof.unfolding ([[(Q_defs, [])]])
  |> Proof.refine_singleton (Method.primitive_text (K I))) ooo
    (fn bisims => fn raw_specs => fn lthy =>
       let
         val specs = snd (fst (Specification.read_multi_specs [] raw_specs lthy))
           |> map (apsnd drop_all);
       in
         gen_nonuninform_coinduct (map (Syntax.read_term lthy) bisims) specs lthy
       end);

val _ = Outer_Syntax.local_theory_to_proof @{command_keyword nonuniform_coinduct}
  "perform coinduction on nonuniform codatatypes"
  (Scan.repeat1 Parse.term --| Parse.$$$ "in"
   -- Parse_Spec.multi_specs >> uncurry nonuninform_coinduct_cmd);

end;
