(*  Title:      HOL/Tools/BNF/bnf_nu_fp_rec.ML
    Author:     Dmitriy Traytel, ETH Zuerich
    Author:     Fabian Meier, ETH Zuerich
    Copyright   2016

(Co)recursive functions over nonuniform (co)datatypes.
*)

signature BNF_NU_FP_REC =
sig
  type nu_lfp_rec_result =
    {f_T_appl: thm}

  val morph_nu_lfp_rec_result: morphism -> nu_lfp_rec_result -> nu_lfp_rec_result

  val construct_nu_rec: mixfix list -> binding list -> BNF_NU_FP_Util.nu_fp_result ->
    BNF_Def.bnf list -> typ list list -> typ list list ->
    BNF_Def.bnf list -> typ list list -> typ list list ->
    ((term * thm) * (term * thm) list list) list -> local_theory -> nu_lfp_rec_result * local_theory
end;

structure BNF_NU_FP_Rec : BNF_NU_FP_REC =
struct

local
  open BNF_Def
  open BNF_Util
  open BNF_Tactics
  open BNF_Comp
  open BNF_FP_Util
  open BNF_FP_Def_Sugar
  open BNF_LFP
  open BNF_LFP_Util
  open BNF_LFP_Tactics
  open BNF_LFP_Rec_Sugar
  open Ctr_Sugar
  open BNF_Tactics
  open BNF_NU_FP
  open BNF_NU_FP_Tactics
in

type nu_lfp_rec_result =
  {
    f_T_appl: thm
  }

fun morph_nu_lfp_rec_result phi {f_T_appl} =
  {
    f_T_appl = Morphism.thm phi f_T_appl
  };

  val mk_b_of_Free = Binding.name o fst o dest_Free;
  val mk_const_val = Const oo pair o fst o dest_Const;
  fun mk_undefined T = Const (@{const_name undefined}, T);
  fun mk_id T = Abs (Name.uu_, T, Bound 0);
  fun mk_predFun_T A B = (A --> HOLogic.boolT) --> (B --> HOLogic.boolT) --> (A --> B) --> HOLogic.boolT;
  fun mk_predFun A B = Const (@{const_name pred_fun}, mk_predFun_T A B);
  val mk_Trueprop_list_conj = HOLogic.mk_Trueprop o Library.foldl1 HOLogic.mk_conj;
  fun pp lthy eq = tap (writeln o Syntax.string_of_term lthy o Syntax.check_term lthy) eq;
  fun pps lthy eqs = map (pp lthy) eqs;
  fun ppss lthy eqss = map (pps lthy) eqss;
  fun pp' lthy eq = tap (writeln o Syntax.string_of_term lthy) eq;
  fun pp's lthy eqs = map (pp' lthy) eqs;
  fun pp'ss lthy eqss = map (pp's lthy) eqss;
  fun flip f a b = f b a;
  fun transp [xs] = map single xs
    | transp (x::xs) = @{map 2} cons x (transp xs)
    | transp [] = [];

  fun Sfn a b = b $ a;
  fun name_of (Const (x, _)) = x
    | name_of (Free (x, _)) = x
    | name_of _ = error "not const or free";

  fun target_ctr_sugar_of_fp_sugar lthy fpT ({T, fp_ctr_sugar = {ctr_sugar, ...}, ...} : fp_sugar) =
    let
      val thy = Proof_Context.theory_of lthy;
      val rho = Vartab.fold (cons o apsnd snd) (Sign.typ_match thy (T, fpT) Vartab.empty) [];
      val phi = Morphism.term_morphism "BNF" (Term.subst_TVars rho);
    in
      morph_ctr_sugar phi ctr_sugar
    end;

  fun construct_nu_rec mixfixes bs nu_fp_res Doms Dom_Dss Css Rans Ran_Dss Bss specs lthy =
  let
    val _ = nu_fp_res : BNF_NU_FP_Util.nu_fp_result
    val nr_occs = map (length o hd o snd) specs;
    val nr_occ = Integer.sum nr_occs;
    val nr_param = (length o snd o hd) specs;
    val nr_Ts = length specs;
    val nr_Bs = live_of_bnf (hd Rans);
    val nr_X = Integer.div_mod nr_Bs nr_param |> fst;
    val nr_Cs = live_of_bnf (hd Doms);

    val _ = if length Rans = nr_Ts then 0 else
      error "Number of Rans must be same as number of Ts";
    val _ = if snd (Integer.div_mod nr_Bs nr_param) = 0 then 0 else
      error "R can only have a number of parameters which is a multiple of the number of parameters of T";
    val _ = if length Doms = nr_param then 0 else
      error "Number of Doms must be same as number of parameters of T";
    val _ = if length (transp Css) = nr_Cs andalso length Css = length Doms then 0 else
      error "Wrong number of type arguments for Doms";
    val _ = if length Bss = nr_X andalso (nr_X = 0 orelse length (hd Bss) = nr_param) then 0 else
      error "Wrong number of type arguments for Rans";

    val dummy_occs = map_index (fn (i, occs) => replicate occs i) nr_occs;
    fun rep_occs xs = @{map 2} replicate nr_occs xs;
    (* Restores occurences out of a list of length nr_Ts *)
    fun rest_occs xs = flat (rep_occs xs);
    (* Restores occurences out of a list of length nr_occ *)
    fun rest_occs2 xs = Library.unflat dummy_occs xs;
    (* Restores occurences out of a nested list of length nr_occ x nr_Ts *)
    fun rest_occs3 xss = rest_occs2 (@{map 2} nth xss (flat dummy_occs))
    fun rep_par x = replicate nr_param x;
    fun rep_occ x = replicate nr_occ x;
    fun rep_X x = replicate nr_X x;
    fun rep_T x = replicate nr_Ts x;

    val defobjs = map (fst o fst) specs;
    val defobj_transfer_thms = map (snd o fst) specs;
    val argobj_infoss = fold (@{map 2} (flip append)) (map snd specs) (rep_par []) |> transpose;
    val argobjss = map (map fst) argobj_infoss;
    val argobj_transfer_thmss = map (map snd) argobj_infoss;

    (* Extract and Create basic types *)

    val As = #As nu_fp_res;
    val Fss = transpose (#Fss nu_fp_res);
    val Dsss = transpose (#Dsss nu_fp_res);
    val Gs = #Gs nu_fp_res;
    val Dss = #Dss nu_fp_res;
    val resDs = #resDs nu_fp_res;
    val shape_sugars = #shape_sugars (#shape_res nu_fp_res) : BNF_FP_Def_Sugar.fp_sugar list;
    val raw_sugars = #raw_sugars nu_fp_res;

    (* Create copies of the base types *)

    val used_letters = fold (union (op =)) ([As] @ Dss @ flat Dsss @ Dom_Dss @ Ran_Dss @ Bss @ Css) resDs;
    val names_lthy = fold Variable.declare_typ used_letters lthy;

    val BCs = fold (union (op =)) ((map single (flat Bss)) @ Css) [];
    val (BC's, _) = mk_TFrees (length BCs) names_lthy;
    val BC_to_BC'_map = BCs ~~ BC's;
    val B'ss = map (map (AList.lookup (op =) BC_to_BC'_map #> the)) Bss;
    val C'ss = map (map (AList.lookup (op =) BC_to_BC'_map #> the)) Css;

    (* All the relevant types and none more... *)
    val D_Cs = @{map 3} mk_T_of_bnf Dom_Dss Css Doms;
    val D_C's = @{map 3} mk_T_of_bnf Dom_Dss C'ss Doms;
    val F_Bsss = @{map 2} (@{map 2} (fn Ds => fn F => map (fn Bs => mk_T_of_bnf Ds Bs F) Bss)) Dsss Fss;
    val F_B'sss = @{map 2} (@{map 2} (fn Ds => fn F => map (fn Bs => mk_T_of_bnf Ds Bs F) B'ss)) Dsss Fss;
    val F_Csss = @{map 2} (@{map 2} (fn Ds => fn F => map (fn Cs => mk_T_of_bnf Ds Cs F) (transp Css))) Dsss Fss;
    val F_C'sss = @{map 2} (@{map 2} (fn Ds => fn F => map (fn Cs => mk_T_of_bnf Ds Cs F) (transp C'ss))) Dsss Fss;
    val F_D_Css = @{map 2} (@{map 2} (fn Ds => mk_T_of_bnf Ds D_Cs)) Dsss Fss;
    val D_F_Css = map (@{map 3} (fn Ds =>fn D => fn Fs => mk_T_of_bnf Ds Fs D) Dom_Dss Doms) F_Csss;
    val R_F_Bs = @{map 3} mk_T_of_bnf (rest_occs Ran_Dss) (map flat F_Bsss) (rest_occs Rans);
    val R_F_B's = @{map 3} mk_T_of_bnf (rest_occs Ran_Dss) (map flat F_B'sss) (rest_occs Rans);
    val G_RD_F_BCs = @{map 2} (fn Ds => mk_T_of_bnf Ds (D_Cs @ R_F_Bs)) Dss Gs;
    val R_Bs = @{map 2} (fn Ds => mk_T_of_bnf Ds (flat (transp Bss))) Ran_Dss Rans;
    val shape_D_Cs = map (mk_T_of_bnf resDs D_Cs o #fp_bnf) shape_sugars;
    val shape_Bss = map (fn S => map (fn Bs => mk_T_of_bnf resDs Bs (#fp_bnf S)) Bss) shape_sugars;
    val shape_Css = map (fn S => map (fn Cs => mk_T_of_bnf resDs Cs (#fp_bnf S)) (transp Css)) shape_sugars;
    val D_shape_Cs = @{map 3} mk_T_of_bnf Dom_Dss shape_Css Doms;
    val F_shape_D_Css = @{map 2} (@{map 2} (fn Ds => mk_T_of_bnf Ds shape_D_Cs)) Dsss Fss;
    val F_shape_Csss = @{map 2} (@{map 2} (fn Ds => fn F => map (fn Ss => mk_T_of_bnf Ds Ss F) (transp shape_Css))) Dsss Fss;
    val raw_D_Cs = map (mk_T_of_bnf resDs D_Cs o #fp_bnf) raw_sugars;
    val R_shape_Bs = @{map 2} (fn Ds => mk_T_of_bnf Ds (flat shape_Bss)) Ran_Dss Rans;
    val F_shape_Bsss = @{map 2} (@{map 2} (fn Ds => fn F => map (fn Ss => mk_T_of_bnf Ds Ss F) (transp shape_Bss))) Dsss Fss;
    val G_raw_D_Cs = @{map 2} (fn Ds => mk_T_of_bnf Ds (shape_D_Cs @ (rest_occs raw_D_Cs))) Dss Gs;
    val R_F_shape_Bs = @{map 3} mk_T_of_bnf (rest_occs Ran_Dss) (map flat F_shape_Bsss) (rest_occs Rans);
    val T_D_Cs = #Ts (#fp_res nu_fp_res)
      |> map (Term.typ_subst_atomic (As ~~ D_Cs));
    val shape_F_D_Css = map (fn Fs => map (mk_T_of_bnf resDs Fs o #fp_bnf) shape_sugars) F_D_Css;
    val shape_F_Csss = map (fn Fss => map (fn sugar => map (fn Fs => mk_T_of_bnf resDs Fs (#fp_bnf sugar)) (transpose Fss)) shape_sugars) F_Csss
    val shape_D_F_Css = map (fn D_Fs => map (mk_T_of_bnf resDs D_Fs o #fp_bnf) shape_sugars) D_F_Css;
    val D_shape_F_Css =
      if nr_Cs <> 0 then
        map (@{map 3} (fn Ds => fn D => fn S_Fs => mk_T_of_bnf Ds S_Fs D) Dom_Dss Doms) shape_F_Csss
      else rep_occ (map2 (fn Ds => mk_T_of_bnf Ds []) Dom_Dss Doms);
    val raw_F_D_Css = map (fn Fs => map (mk_T_of_bnf resDs Fs o #fp_bnf) raw_sugars) F_D_Css;
    val shape_F_Bsss = map (fn Fss => map (fn sugar => map (fn Fs => mk_T_of_bnf resDs Fs (#fp_bnf sugar)) (transpose Fss)) shape_sugars) F_Bsss
    val raw_D_F_Css = map (fn D_Fs => map (mk_T_of_bnf resDs D_Fs o #fp_bnf) raw_sugars) D_F_Css;
    val R_shape_F_Bss =
      if nr_X <> 0 then
        map (fn S_Fss => @{map 2} (fn Ds => mk_T_of_bnf Ds (flat S_Fss)) Ran_Dss Rans) shape_F_Bsss
      else
        rep_occ (map2 (fn Ds => mk_T_of_bnf Ds []) Ran_Dss Rans);
    val T_D_F_Cs = #Ts (#fp_res nu_fp_res)
      |> @{map 2} (fn Ds => Term.typ_subst_atomic (As ~~ Ds)) D_F_Css o rest_occs;
    val T_F_D_Cs = #Ts (#fp_res nu_fp_res)
      |> @{map 2} (fn Fs => Term.typ_subst_atomic (As ~~ Fs)) F_D_Css o rest_occs;
    val G_T_F_D_Cs = @{map 2} (fn Ds => mk_T_of_bnf Ds (D_Cs @ T_F_D_Cs)) Dss Gs;
    val Depth = (HOLogic.listT o #T o #label_sugar o #shape_res) nu_fp_res;

    (* Create Frees and Consts *)

    val (((((((((((fnB_valss, fnC_valss), grdf_vals), fd_valss), pdB_valss), pdC_valss),
      d_vals), fsd_valss), shape_valss), grd_vals), td_vals), _) = lthy
      |> mk_Freess "fnB" (@{map 2} (@{map 2} (curry op -->)) Bss B'ss)
      ||>> mk_Freess "fnC" (@{map 2} (@{map 2} (curry op -->)) Css C'ss)
      ||>> mk_Frees "grdf" G_RD_F_BCs
      ||>> mk_Freess "fd" F_D_Css
      ||>> mk_Freess "pdB" (map (map (fn B => B --> HOLogic.boolT)) Bss)
      ||>> mk_Freess "pdC" (map (map (fn C => C --> HOLogic.boolT)) Css)
      ||>> mk_Frees "d" D_Cs
      ||>> mk_Freess "fsd" F_shape_D_Css
      ||>> mk_Freess "sb" shape_Bss
      ||>> mk_Frees "grd" G_raw_D_Cs
      ||>> mk_Frees "td" T_D_Cs;

    val BC_to_shapeBC_map = (flat Bss ~~ flat shape_Bss) @ (flat Css ~~ flat shape_Css);
    val defobj_vals = defobjs;
    val defobj'_vals = map (Term.subst_atomic_types BC_to_BC'_map) defobjs;
    val defobj_shape_vals = map (Term.subst_atomic_types BC_to_shapeBC_map) defobjs;
    val argobj_valss = argobjss;
    val argobj'_valss = map (map (Term.subst_atomic_types BC_to_BC'_map)) argobjss;
    val argobj_shape_valss = map (map (Term.subst_atomic_types BC_to_shapeBC_map)) argobjss;
    val invar_shape_vals = #invar_shape_vals (#shape_res nu_fp_res);
    val flat_shape_valss = #flat_shape_valss (#shape_res nu_fp_res);

    (* defobj natural thms *)

    val mapDs = @{map 4} mk_map_of_bnf Dom_Dss Css C'ss Doms;
    val mapR_Fs = @{map 4} mk_map_of_bnf (rest_occs Ran_Dss) (map flat F_Bsss) (map flat F_B'sss) (rest_occs Rans);
    val mapF_Bsss = @{map 2} (@{map 2} (fn Ds => fn F => @{map 2} (fn Bs => fn B's => mk_map_of_bnf Ds Bs B's F) Bss B'ss)) Dsss Fss;
    val mapGs = @{map 2} (fn Ds => mk_map_of_bnf Ds (D_Cs @ R_F_Bs) (D_C's @ R_F_B's)) Dss Gs;
    val mapRs = @{map 2} (fn Ds => mk_map_of_bnf Ds (flat (transp Bss)) (flat (transp B'ss))) Ran_Dss Rans;

    fun mk_defobj_natural mapDs fCss mapRs mapFsss fBss defobj' mapG g mapR defobj =
      mk_Trueprop_eq (defobj' $ (list_comb (list_comb (mapG,
      map list_comb (mapDs ~~ fCss)),
      map list_comb (mapRs ~~
      map (flat o map (@{map 2} (list_comb oo rpair) fBss)) mapFsss)) $ g),
      list_comb (mapR, flat (transp fBss)) $ (defobj $ g));
    val defobj_natural_goals = @{map 5} (mk_defobj_natural mapDs fnC_valss mapR_Fs mapF_Bsss fnB_valss)
      defobj'_vals mapGs grdf_vals mapRs defobj_vals
|> pps lthy;
    val defobj_natural_varss = map (fn g => map name_of (g :: (flat (fnB_valss @ fnC_valss)))) grdf_vals;

    val defobj_natural_tactic =
      (fn {context = ctxt, prems = _} =>
        BNF_NU_FP_Tactics.mk_defobj_natural_tac ctxt);

    val defobj_natural_thms = @{map 2} (fn goal => fn vars =>
          Goal.prove_sorry lthy vars [] goal defobj_natural_tactic
          |> Thm.close_derivation)
          defobj_natural_goals defobj_natural_varss;

    (* argobj natural thms *)

    val mapF_Csss = @{map 2} (@{map 2} (fn Ds => fn F => @{map 2} (fn Cs => fn C's => mk_map_of_bnf Ds Cs C's F) (transp Css) (transp C'ss))) Dsss Fss;
    val mapF_Dss = @{map 2} (@{map 2} (fn Ds => mk_map_of_bnf Ds D_Cs D_C's)) Dsss Fss;
    val mapD_Fss = @{map 2} (@{map 4} (fn Ds => fn D => fn F_Cs => fn F_C's => mk_map_of_bnf Ds F_Cs F_C's D) Dom_Dss Doms) F_Csss F_C'sss;

    fun mk_argobj_natural mapDs fCss  argobj' mapF_D f mapD mapFs argobj =
      mk_Trueprop_eq (argobj' $ (list_comb (mapF_D,
      map list_comb (mapDs ~~ fCss)) $ f),
      list_comb (mapD, map list_comb (mapFs ~~ transp fCss)) $ (argobj $ f));
    val argobj_natural_goalss = @{map 6} (@{map 6} (mk_argobj_natural mapDs fnC_valss))
      argobj'_valss mapF_Dss fd_valss mapD_Fss mapF_Csss argobj_valss
|> ppss lthy;
    val argobj_natural_varsss = map (map (fn fd => map name_of (fd :: (flat fnC_valss)))) fd_valss;

    val argobj_natural_tactic =
      (fn {context = ctxt, prems = _} =>
        BNF_NU_FP_Tactics.mk_argobj_natural_tac ctxt);

    val argobj_natural_thmss = @{map 2} (@{map 2} (fn goal => fn vars =>
          Goal.prove_sorry lthy vars [] goal argobj_natural_tactic
          |> Thm.close_derivation))
          argobj_natural_goalss argobj_natural_varsss;

    (* defobj invar thms *)

    val predD_Cs = @{map 3} mk_pred_of_bnf Dom_Dss Css Doms;
    val predR_Fs = @{map 3} mk_pred_of_bnf (rest_occs Ran_Dss) (map flat F_Bsss) (rest_occs Rans);
    val predF_Bsss = @{map 2} (@{map 2} (fn Ds => fn F => map (fn Bs => mk_pred_of_bnf Ds Bs F) Bss)) Dsss Fss;
    val predFun_vals = @{map 2} mk_predFun G_RD_F_BCs R_Bs;
    val predGs = @{map 2} (fn Ds => mk_pred_of_bnf Ds (D_Cs @ R_F_Bs)) Dss Gs;
    val predRs = @{map 2} (fn Ds => mk_pred_of_bnf Ds (flat (transp Bss))) Ran_Dss Rans;

    fun mk_defobj_invar predDs pCss predRs predFsss pBss predFun predG predR defobj =
      HOLogic.mk_Trueprop (predFun $ list_comb (list_comb (predG,
      map list_comb (predDs ~~ pCss)),
      map list_comb (predRs ~~
      map (flat o map (@{map 2} (list_comb oo rpair) pBss)) predFsss)) $
      list_comb (predR, flat (transp pBss)) $ defobj);
    val defobj_invar_goals = @{map 4} (mk_defobj_invar predD_Cs pdC_valss predR_Fs predF_Bsss pdB_valss)
      predFun_vals predGs predRs defobj_vals
|> pps lthy;
    val defobj_invar_vars = (map name_of o flat) (pdC_valss @ pdB_valss);

    val defobj_invar_tactic =
      (fn {context = ctxt, prems = _} =>
        BNF_NU_FP_Tactics.mk_defobj_invar_tac ctxt);

    val defobj_invar_thms = map (fn goal =>
          Goal.prove_sorry lthy defobj_invar_vars [] goal defobj_invar_tactic
          |> Thm.close_derivation)
          defobj_invar_goals;

    (* argobj invar thms *)

    val predFun_valss = @{map 2} (@{map 2} mk_predFun) F_D_Css D_F_Css;
    val predF_Dss = @{map 2} (@{map 2} (fn Ds => mk_pred_of_bnf Ds D_Cs)) Dsss Fss;
    val predD_Fss = map (@{map 3} (fn Ds => fn D => fn F_Cs => mk_pred_of_bnf Ds F_Cs D) Dom_Dss Doms) F_Csss;
    val predF_Csss = @{map 2} (@{map 2} (fn Ds => fn F => map (fn Cs => mk_pred_of_bnf Ds Cs F) (transp Css))) Dsss Fss;

    fun mk_argobj_invar predDs pCss predFun predF predD predFs argobj =
      HOLogic.mk_Trueprop (predFun $ list_comb (predF,
      map list_comb (predDs ~~ pCss)) $
      list_comb (predD, map list_comb (predFs ~~ transp pCss)) $ argobj);

    val argobj_invar_goalss = @{map 5} (@{map 5} (mk_argobj_invar predD_Cs pdC_valss))
      predFun_valss predF_Dss predD_Fss predF_Csss argobj_valss
|> ppss lthy;
    val argobj_invar_vars = (map name_of o flat) (pdC_valss);

    val argobj_invar_tactic =
      (fn {context = ctxt, prems = _} =>
        BNF_NU_FP_Tactics.mk_argobj_invar_tac ctxt);

    val argobj_invar_thmss = map (map (fn goal =>
          Goal.prove_sorry lthy argobj_invar_vars [] goal argobj_invar_tactic
          |> Thm.close_derivation))
          argobj_invar_goalss;

    (* f_shape *)

    val f_shape_Ts = @{map 2} (curry op -->) shape_D_Cs D_shape_Cs;
    val (f_shapes_, _) = mk_Frees "f_shape" f_shape_Ts lthy;
    val shape_D_ctrss = @{map 2} (#ctrs oo target_ctr_sugar_of_fp_sugar lthy)
      shape_D_Cs shape_sugars;
    val shape_ctrsss = @{map 2} (fn sugar => map (fn S => target_ctr_sugar_of_fp_sugar lthy S sugar |> #ctrs))
      shape_sugars shape_Css;
    val mapD_shapes = @{map 4} mk_map_of_bnf Dom_Dss Css shape_Css Doms;
    val mapD_Fss = map (@{map 4} (fn Ds => fn Ss => fn D => fn Fs => (mk_map_of_bnf Ds Fs Ss D)) Dom_Dss shape_Css Doms) F_shape_Csss;
    val mapF_shapess = @{map 2} (@{map 2} (fn Ds => mk_map_of_bnf Ds shape_D_Cs D_shape_Cs)) Dsss Fss;

    fun mk_f_shape_Leaf_eq f_shape leaf_D a mapD leafs =
      mk_Trueprop_eq (f_shape $ (leaf_D $ a), list_comb (mapD, leafs) $ a);
    val f_Leaf_eqs = @{map 5} mk_f_shape_Leaf_eq
      f_shapes_ (map hd shape_D_ctrss) d_vals mapD_shapes (map (map hd) shape_ctrsss)
|> pps lthy;

    fun mk_f_shape_Node_eq f_shapes f_shape node_D f mapD nodes argobj mapF =
      mk_Trueprop_eq (f_shape $ (node_D $ f),
      list_comb (mapD, nodes) $ (argobj $ (list_comb (mapF, f_shapes) $ f)));
    val shape_Nodesss = if nr_Cs <> 0 then (map (map tl) shape_ctrsss |> map transpose |> transpose) else rep_occ (rep_par []);
    val f_Node_eqss = @{map 6} (@{map 7} (mk_f_shape_Node_eq f_shapes_) f_shapes_)
       (map tl shape_D_ctrss |> transpose) fsd_valss mapD_Fss shape_Nodesss argobj_shape_valss mapF_shapess
|> ppss lthy;

    val f_shape_eqs = f_Leaf_eqs @ flat f_Node_eqss;
    val f_shape_decls = map (fn f_shape => (mk_b_of_Free f_shape, NONE, NoSyn)) f_shapes_;

    val ((f_shape_free, _, f_shape_simps_free), (lthy, lthy_old)) = lthy
      |> Local_Theory.open_target |> snd
      |> primrec false [] f_shape_decls
        (map (fn eq => ((Binding.empty_atts, eq), [], [])) f_shape_eqs)
      ||> `Local_Theory.close_target;

    val phi = Proof_Context.export_morphism lthy_old lthy;
    val f_shapes = map (Morphism.term phi) f_shape_free;
    val f_shape_simps = map (Morphism.fact phi) f_shape_simps_free |> flat;
    val f_shape_vals = @{map 2} mk_const_val f_shapes f_shape_Ts;

    (* f_raw *)

    val f_raw_Ts = @{map 2} (curry op -->) raw_D_Cs R_shape_Bs;
    val (f_raws_, _) = mk_Frees "f_raw" f_raw_Ts lthy;
    val mapR_Fs = @{map 3} (fn Ds => mk_map_of_bnf Ds (flat shape_Bss)) (rest_occs Ran_Dss) (map flat F_shape_Bsss) (rest_occs Rans);
    val unNodesss = if nr_X <> 0 then
      let
        val raw_case_trmsss = map (@{map 2} (fn sugar =>
          @{map 2} (fn Bs => fn T =>
          mk_case (resDs @ Bs) T (#casex (#ctr_sugar (#fp_ctr_sugar sugar))))
          Bss) shape_sugars) F_shape_Bsss;
        val undefsss = map (map (map mk_undefined)) F_shape_Bsss;

        fun mk_case_trm i B Fs case_trm undef =
          list_comb (case_trm $ (absdummy B undef), @{map 2} (fn F => fn j =>
          if i = j then mk_id F else absdummy F undef) Fs (1 upto nr_occ));

        val case_trmsss = @{map 3} (fn i => @{map 4} (@{map 4} (mk_case_trm i))
          (transp Bss) (map transp (transp F_shape_Bsss)))
          (1 upto nr_occ) raw_case_trmsss undefsss;
      in
        map (@{map 2} (@{map 2} (fn s => fn c => lambda s (c $ s))) shape_valss) case_trmsss
      end else rep_occ [];
    val raw_D_ctrs = @{map 2} (hd o #ctrs oo target_ctr_sugar_of_fp_sugar lthy) raw_D_Cs raw_sugars;
    val mapG_Ss =  @{map 2} (fn Ds => mk_map_of_bnf Ds (shape_D_Cs @ (rest_occs raw_D_Cs)) (D_shape_Cs @ R_F_shape_Bs)) Dss Gs

    fun mk_f_raw f_shapes mapRs unNodesss f_raws R_shapes  f_raw cons g defobj mapG =
      mk_Trueprop_eq (f_raw $ (cons $ g),
      defobj $ (list_comb (list_comb (mapG, f_shapes),
      @{map 4} (fn mapR => fn unNodess => fn f_raw => fn R_shape =>
      mk_comp [R_shape] (list_comb (mapR, flat unNodess), f_raw))
      mapRs unNodesss f_raws R_shapes) $ g));

    val f_raw_eqs = @{map 5} (mk_f_raw f_shape_vals mapR_Fs unNodesss (rest_occs f_raws_) (rest_occs R_shape_Bs))
      f_raws_ raw_D_ctrs grd_vals defobj_shape_vals mapG_Ss
|> pps lthy;
    val f_raw_decls = map (fn f_raw => (mk_b_of_Free f_raw, NONE, NoSyn)) f_raws_;

    val ((f_raw_free, _, f_raw_simps_free), (lthy, lthy_old)) = lthy
      |> Local_Theory.open_target |> snd
      |> primrec false [] f_raw_decls (map (fn eq => ((Binding.empty_atts, eq), [], [])) f_raw_eqs)
      ||> `Local_Theory.close_target;

    val phi = Proof_Context.export_morphism lthy_old lthy;
    val f_raws = map (Morphism.term phi) f_raw_free;
    val f_raw_simps = map (Morphism.fact phi) f_raw_simps_free |> flat;
    val f_raw_vals = @{map 2} mk_const_val f_raws f_raw_Ts;

    (* f_T *)

    val f_T_Ts = @{map 2} (curry op -->) T_D_Cs R_Bs;
    val (f_T_, _) = mk_Frees "f_T" f_T_Ts lthy;
    val unLeavess =
      if nr_X <> 0 then
        let
          val raw_case_trmss = @{map 2} (fn sugar => @{map 2} (fn Bs => fn T =>
            ((mk_case (resDs @ Bs) T) o #casex o #ctr_sugar o #fp_ctr_sugar) sugar)
            Bss) shape_sugars (transp Bss);
          val undefss = map (map mk_undefined) (transp Bss);

          fun mk_case_trm A Fs case_trm undef =
            list_comb (case_trm $ (mk_id A), map (flip absdummy undef) Fs);
          val case_trmss = @{map 4} (@{map 4} mk_case_trm)
            (transpose Bss) (map transp (transp F_shape_Bsss)) raw_case_trmss undefss;
        in
          @{map 2} (@{map 2} (fn s => fn c => lambda s (c $ s))) shape_valss case_trmss
        end
      else [];
    val mapR_Ss = @{map 2} (fn Ds => mk_map_of_bnf Ds (flat shape_Bss) (flat (transp Bss))) Ran_Dss Rans;
    val T_Rep_Ts = @{map 2} (curry op -->) T_D_Cs raw_D_Cs;
    val T_Rep_vals = @{map 2} (Const oo pair o #Rep_name o fst) (#T_infos nu_fp_res) T_Rep_Ts;

    fun mk_f_T unLeavess t mapR f_raw T_Rep =
      lambda t (list_comb (mapR, flat unLeavess) $ (f_raw $ (T_Rep $ t)));
    val f_T_eqs = @{map 4} (mk_f_T unLeavess) td_vals mapR_Ss f_raw_vals T_Rep_vals
|> pps lthy;
    val f_T_decls = map (rpair NoSyn o mk_b_of_Free) f_T_;

    val ((f_Ts, f_T_defs), lthy) =
      @{fold 2} (fn f_T_eq => fn f_T_decl => fn (res, lthy) =>
      let
        val ((f_T_free, (_, f_T_def_free)), (lthy, lthy_old)) =
          lthy
          |> Local_Theory.open_target |> snd
          |> Local_Theory.define (f_T_decl, (Binding.empty_atts, f_T_eq))
          ||> `Local_Theory.close_target;

        val phi = Proof_Context.export_morphism lthy_old lthy;
        val f_T = Morphism.term phi f_T_free;
        val f_T_def = Morphism.thm phi f_T_def_free;
      in (res @ [(f_T, f_T_def)], lthy) end)
      f_T_eqs f_T_decls ([], lthy)
      |> apfst @{split_list 2};
    val f_T_vals = @{map 2} mk_const_val f_Ts f_T_Ts;

    (* Lemmas about f *)

    val (((((sfd_valss, sd_vals), rfd_valss), gtfd_vals), u0), _) = lthy
      |> mk_Freess "sfd" shape_F_D_Css
      ||>> mk_Frees "sd" shape_D_Cs
      ||>> mk_Freess "rfd" raw_F_D_Css
      ||>> mk_Frees "gtfd" G_T_F_D_Cs
      ||>> yield_singleton (mk_Frees "u0") Depth;

    (* invar_shape_f_shape *)

    val invar_shape_Tss = map (map (fn shape => Depth --> shape --> HOLogic.boolT)) shape_Css;
    val invar_shape_valss = @{map 2} (fn const => map (mk_const_val const)) invar_shape_vals invar_shape_Tss;
    val invar_shape_D_Ts = map (fn S_D => Depth --> S_D --> HOLogic.boolT) shape_D_Cs;
    val invar_shape_D_vals = @{map 2} mk_const_val invar_shape_vals invar_shape_D_Ts;
    val predDs = @{map 3} mk_pred_of_bnf Dom_Dss shape_Css Doms;

    fun mk_invar_shape_f_shape invar_D u predD invars f_shape =
      lambda u (HOLogic.mk_all (name_of u0, fastype_of u0,
      HOLogic.mk_imp (invar_D $ u0 $ u,
      list_comb (predD, map (Sfn u0) invars) $ (f_shape $ u))));

    val invar_shape_f_shape_subgoals = @{map 5} mk_invar_shape_f_shape
      invar_shape_D_vals sd_vals predDs invar_shape_valss f_shape_vals;
    val invar_shape_f_shape_goals = map2 Sfn sd_vals invar_shape_f_shape_subgoals;
    val invar_shape_f_shape_goal = mk_Trueprop_list_conj invar_shape_f_shape_goals;
    val invar_shape_f_shape_vars = map name_of sd_vals;

    val invar_shape_f_shape_tactic =
      (fn {context = ctxt, prems = _} =>
        BNF_NU_FP_Tactics.mk_invar_shape_f_shape_tac ctxt);

    val invar_shape_f_shape_thm =
      Goal.prove_sorry lthy invar_shape_f_shape_vars [] invar_shape_f_shape_goal invar_shape_f_shape_tactic
      |> Thm.close_derivation
|>  (tap (writeln o Thm.string_of_thm lthy));

    (* f_shape_flat_shape *)

    val flat_shape_Tsss =
      if nr_Cs <> 0 then
        map (@{map 2} (@{map 2} (flip (curry op -->))) shape_Css) shape_F_Csss
      else
        rep_occ (rep_par []);
    val flat_shape_valsss = @{map 2} (@{map 2} (fn const => map (mk_const_val const))) flat_shape_valss flat_shape_Tsss;
    val flat_shape_D_Tss = map (@{map 2} (flip (curry op -->)) shape_D_Cs) shape_F_D_Css;
    val flat_shape_D_valss = @{map 2} (@{map 2} mk_const_val) flat_shape_valss flat_shape_D_Tss;
    val invar_shape_F_D_Tss = map (map (fn shape => Depth --> shape --> HOLogic.boolT)) shape_F_D_Css;
    val invar_shape_F_D_valss = map (@{map 2} mk_const_val invar_shape_vals) invar_shape_F_D_Tss;
    val f_shape_F_Tss = @{map 2} (@{map 2} (curry op -->)) shape_D_F_Css D_shape_F_Css;
    val f_shape_F_valss = map (@{map 2} mk_const_val f_shape_vals) f_shape_F_Tss;
    val mapD_Sss =
      if nr_Cs <> 0 then
        map (@{map 4} (fn Ds => fn Ss => fn D => fn S_Fs => mk_map_of_bnf Ds S_Fs Ss D) Dom_Dss shape_Css Doms) shape_F_Csss
      else
        rep_occ (map2 (fn Ds => mk_map_of_bnf Ds [] []) Dom_Dss Doms);
    val mapS_Fss = @{map 2} (fn F_Ds => fn D_Fs => map (mk_map_of_bnf resDs F_Ds D_Fs o #fp_bnf) shape_sugars) F_D_Css D_F_Css;

    fun mk_f_shape_flat_shape argobjs f_shape invar flat_shape_D u mapD flat_shapes f_shape_F mapS =
      lambda u (HOLogic.mk_all (name_of u0, fastype_of u0,
      HOLogic.mk_imp (invar $ u0 $ u,
      HOLogic.mk_eq (f_shape $ (flat_shape_D $ u),
      list_comb (mapD, flat_shapes) $ (f_shape_F $ (list_comb (mapS, argobjs) $ u))))));

    val f_shape_flat_shape_subgoalss = @{map 8} (fn argobjs => @{map 8} (mk_f_shape_flat_shape argobjs)
      f_shape_vals) argobj_valss invar_shape_F_D_valss flat_shape_D_valss sfd_valss mapD_Sss flat_shape_valsss f_shape_F_valss mapS_Fss;
    val f_shape_flat_shape_goalss = @{map 2} (@{map 2} Sfn) sfd_valss f_shape_flat_shape_subgoalss;
    val f_shape_flat_shape_goals = map mk_Trueprop_list_conj f_shape_flat_shape_goalss;
    val f_shape_flat_shape_varss = map (map name_of) sfd_valss;

    val f_shape_flat_shape_tactic =
      (fn {context = ctxt, prems = _} =>
        BNF_NU_FP_Tactics.mk_f_shape_flat_shape_tac ctxt);

    val f_shape_flat_shape_thms = @{map 2} (fn goal => fn vars =>
          Goal.prove_sorry lthy vars [] goal f_shape_flat_shape_tactic
          |> Thm.close_derivation)
          f_shape_flat_shape_goals f_shape_flat_shape_varss
|> map (tap (writeln o Thm.string_of_thm lthy));

    (* invar_raw_f_raw *)

    val invar_shape_Tsss = map (map (map (fn shape => Depth --> shape --> HOLogic.boolT))) shape_F_Bsss;
    val invar_shape_valsss = map (@{map 2} (fn const => map (mk_const_val const)) invar_shape_vals) invar_shape_Tsss;
    val invar_raw_D_Tss = map (map (fn raw => Depth --> raw --> HOLogic.boolT)) raw_F_D_Css;
    val invar_raw_D_valss = map (@{map 2} mk_const_val (#invar_raws nu_fp_res)) invar_raw_D_Tss;
    val predR_S_Fss =
      if nr_X <> 0 then
        map (fn S_Fss => @{map 2} (fn Ds => mk_pred_of_bnf Ds (flat S_Fss)) Ran_Dss Rans) shape_F_Bsss
      else
        rep_occ (map2 (fn Ds => mk_pred_of_bnf Ds []) Ran_Dss Rans);
    val f_raw_F_Tss = map2 (@{map 2} (curry op -->)) raw_D_F_Css R_shape_F_Bss;
    val f_raw_F_valss = map (@{map 2} mk_const_val f_raw_vals) f_raw_F_Tss;
    val map_rawss = @{map 2} (fn F_Ds => fn D_Fs =>
      map (mk_map_of_bnf resDs F_Ds D_Fs o #fp_bnf) raw_sugars) F_D_Css D_F_Css;

    fun mk_invar_raw_f_raw argobjs invar_shapess invar_D predR t f_raw map_raw =
      lambda t (HOLogic.mk_all (name_of u0, fastype_of u0,
      HOLogic.mk_imp (invar_D $ u0 $ t,
      list_comb (predR, map (Sfn u0) (flat invar_shapess)) $
      (f_raw $ (list_comb (map_raw, argobjs) $ t)))));

    val invar_raw_f_raw_subgoalss = @{map 7} (fn argobjs => fn invar_shapes =>
      @{map 5} (mk_invar_raw_f_raw argobjs invar_shapes)) argobj_valss invar_shape_valsss invar_raw_D_valss
      predR_S_Fss rfd_valss f_raw_F_valss map_rawss;
    val invar_raw_f_raw_goalss = @{map 2} (@{map 2} Sfn) rfd_valss invar_raw_f_raw_subgoalss;
    val invar_raw_f_raw_goals = map mk_Trueprop_list_conj invar_raw_f_raw_goalss;
    val invar_raw_f_raw_varss = map (map name_of) rfd_valss;

    val invar_raw_f_raw_tactic =
      (fn {context = ctxt, prems = _} =>
        BNF_NU_FP_Tactics.mk_invar_raw_f_raw_tac ctxt);

    val invar_raw_f_raw_thms = @{map 2} (fn goal => fn vars =>
          Goal.prove_sorry lthy vars [] goal invar_raw_f_raw_tactic
          |> Thm.close_derivation)
          invar_raw_f_raw_goals invar_raw_f_raw_varss
|> map (tap (writeln o Thm.string_of_thm lthy));

    (* f_raw_flat_raw *)

    val flat_shape_Tsss =
      if nr_X <> 0 then
        map (@{map 2} (@{map 2} (flip (curry op -->))) shape_Bss) shape_F_Bsss
      else rep_occ (rep_par []);
    val flat_shape_valsss = @{map 2} (@{map 2} (fn const => map (mk_const_val const))) flat_shape_valss flat_shape_Tsss;
    val flat_raw_D_Tss = map (@{map 2} (flip (curry op -->)) raw_D_Cs) raw_F_D_Css;
    val flat_raw_D_valss = @{map 2} (@{map 2} mk_const_val) (#flat_rawss nu_fp_res) flat_raw_D_Tss;
    val mapR_Sss =
      if nr_X <> 0 then
        map (fn S_Fs => @{map 2} (fn Ds => mk_map_of_bnf Ds (flat S_Fs) (flat shape_Bss)) Ran_Dss Rans) shape_F_Bsss
      else
        rep_occ (map2 (fn Ds => mk_map_of_bnf Ds [] []) Ran_Dss Rans);
    val f_raw_F_Tss = @{map 2} (@{map 2} (curry op -->)) raw_D_F_Css R_shape_F_Bss;
    val f_raw_F_valss = map (@{map 2} mk_const_val f_raw_vals) f_raw_F_Tss;

    fun mk_f_raw_flat_raw flatss argobjs f_raw invar_D t flat_D mapR f_raw_F mapr =
      lambda t (HOLogic.mk_all (name_of u0, fastype_of u0,
      HOLogic.mk_imp (invar_D $ u0 $ t,
      HOLogic.mk_eq (f_raw $ (flat_D $ t),
      list_comb (mapR, flat flatss) $ (f_raw_F $ (list_comb (mapr, argobjs) $ t))))));

    val f_raw_flat_raw_subgoalss = @{map 8} (fn flatss => fn argobjs =>
      @{map 7} (mk_f_raw_flat_raw flatss argobjs) f_raw_vals) flat_shape_valsss argobj_valss
      invar_raw_D_valss rfd_valss flat_raw_D_valss mapR_Sss f_raw_F_valss map_rawss;
    val f_raw_flat_raw_goalss = @{map 2} (@{map 2} Sfn) rfd_valss f_raw_flat_raw_subgoalss;
    val f_raw_flat_raw_goals = map mk_Trueprop_list_conj f_raw_flat_raw_goalss
    val f_raw_flat_raw_varss = map (map name_of) rfd_valss;

    val f_raw_flat_raw_tactic =
      (fn {context = ctxt, prems = _} =>
        BNF_NU_FP_Tactics.mk_f_raw_flat_raw_tac ctxt);

    val f_raw_flat_raw_thms = @{map 2} (fn goal => fn vars =>
          Goal.prove_sorry lthy vars [] goal f_raw_flat_raw_tactic
          |> Thm.close_derivation)
          f_raw_flat_raw_goals f_raw_flat_raw_varss
      |> map (tap (writeln o Thm.string_of_thm lthy));

    (* f_T_Appl *)

    val idDs = map mk_id D_Cs;
    val f_T_F_Ts = @{map 2} (curry op -->) T_D_F_Cs R_F_Bs;
    val f_T_F_vals = @{map 2} mk_const_val (rest_occs f_Ts) f_T_F_Ts;
    val mapTs = @{map 3} (mk_map_of_bnf resDs) F_D_Css D_F_Css ((rest_occs o #bnfs o #fp_res) nu_fp_res);
    val T_ctr_Ts = @{map 2} (curry op -->) G_T_F_D_Cs T_D_Cs;
    val T_ctr_vals = @{map 2} mk_const_val ((#ctors o #fp_res) nu_fp_res) T_ctr_Ts;
    val mapG_Ts = @{map 2} (fn Ds => mk_map_of_bnf Ds (D_Cs @ T_F_D_Cs) (D_Cs @ R_F_Bs)) Dss Gs;

    fun mk_f_T_Appl ids f_Ts mapTs argobjss f_T T_ctr g defobj mapG =
      lambda g (HOLogic.mk_eq (f_T $ (T_ctr $ g),
      defobj $ (list_comb(list_comb (mapG, ids),
      @{map 3} (fn f_T => fn mapT => fn argobjs =>
      HOLogic.mk_comp (f_T, list_comb (mapT, argobjs)))
      f_Ts mapTs argobjss) $ g)));
    val f_T_Appl_subgoals = @{map 5} (mk_f_T_Appl idDs f_T_F_vals mapTs argobj_valss)
      f_T_vals T_ctr_vals gtfd_vals defobj_vals mapG_Ts;
    val f_T_Appl_goals = @{map 2} Sfn gtfd_vals f_T_Appl_subgoals;
    val f_T_Appl_goal = mk_Trueprop_list_conj f_T_Appl_goals;
    val f_T_Appl_vars = map name_of gtfd_vals;

    val f_T_Appl_tactic =
      (fn {context = ctxt, prems = _} =>
        BNF_NU_FP_Tactics.mk_f_T_Appl_tac ctxt);

    val f_T_Appl_thm =
          Goal.prove_sorry lthy f_T_Appl_vars [] f_T_Appl_goal f_T_Appl_tactic
          |> Thm.close_derivation
|>  (tap (writeln o Thm.string_of_thm lthy));

    val nu_lfp_rec_res = {
        f_T_appl = f_T_Appl_thm
      };
  in
    (nu_lfp_rec_res, lthy)
  end;

end;

end;
