(*  Title:      HOL/Tools/BNF/bnf_nu_fp_rec_sugar.ML
    Author:     Jasmin Blanchette, Inria, LORIA, MPII
    Copyright   2016

Sugared (co)recursion over nonuniform (co)datatypes.
*)

signature BNF_NU_FP_REC_SUGAR =
sig
  datatype nu_rec_option =
    Plugins_NU_Rec_Option of Proof.context -> Plugin_Name.filter

  val nonuniform_primrec: bool -> nu_rec_option list -> (binding * typ option * mixfix) list ->
    Specification.multi_specs -> local_theory -> thm list * local_theory
  val nonuniform_primrec_cmd: bool -> nu_rec_option list ->
    (binding * string option * mixfix) list -> Specification.multi_specs_cmd -> local_theory ->
    thm list * local_theory
  val nonuniform_primrecursive: bool -> nu_rec_option list ->
    (binding * typ option * mixfix) list -> Specification.multi_specs -> local_theory ->
    Proof.state
  val nonuniform_primrecursive_cmd: bool -> nu_rec_option list ->
    (binding * string option * mixfix) list -> Specification.multi_specs_cmd -> local_theory ->
    Proof.state

  datatype nu_corec_option =
    Plugins_NU_Corec_Option of Proof.context -> Plugin_Name.filter

  val nonuniform_primcorec: bool -> nu_corec_option list -> (binding * typ option * mixfix) list ->
    Specification.multi_specs -> local_theory -> thm list * local_theory
  val nonuniform_primcorec_cmd: bool -> nu_corec_option list ->
    (binding * string option * mixfix) list -> Specification.multi_specs_cmd -> local_theory ->
    thm list * local_theory
  val nonuniform_primcorecursive: bool -> nu_corec_option list ->
    (binding * typ option * mixfix) list -> Specification.multi_specs -> local_theory ->
    Proof.state
  val nonuniform_primcorecursive_cmd: bool -> nu_corec_option list ->
    (binding * string option * mixfix) list -> Specification.multi_specs_cmd -> local_theory ->
    Proof.state
end;

structure BNF_NU_FP_Rec_Sugar : BNF_NU_FP_REC_SUGAR =
struct

open Ctr_Sugar
open BNF_FP_Rec_Sugar_Util

val simpsN = "simps";

(* FIXME: factor out from "primrec" etc. *)
fun malformed_eqn_lhs_rhs ctxt eqn =
  error_at ctxt [eqn] "Malformed equation (expected \"lhs = rhs\")";
fun malformed_eqn_head ctxt eqn =
  error_at ctxt [eqn] "Malformed function equation (expected function name on left-hand side)";

datatype nu_rec_option =
  Plugins_NU_Rec_Option of Proof.context -> Plugin_Name.filter;

fun gen_nonuniform_primrec_ursive prep_spec int long_cmd opts raw_fixes raw_specs lthy =
  let
    val dups = duplicates (op =) (map (Binding.name_of o #1) raw_fixes);
    val _ = null dups orelse error ("Duplicate function name " ^ quote (hd dups));

    val plugins = get_first (fn Plugins_NU_Rec_Option f => SOME (f lthy) | _ => NONE) (rev opts)
      |> the_default Plugin_Name.default_filter;

    val code_attrs =
      if plugins code_plugin then [Code.add_default_eqn_attrib Code.Equation] else [];

    val ((fixes, specs), _) = prep_spec raw_fixes raw_specs lthy;

    val (bs, mxs) = map_split (apfst fst) fixes;
    val frees = map (fst #>> Binding.name_of #> Free) fixes;
    val fun_names = map Binding.name_of bs;
    val qualifys = map (fold_rev (uncurry Binding.qualify o swap) o Binding.path_of) bs;

    (* FIXME: there should be just one equation per type *)

    fun head_fun_name_of_eqn
          (eqn as @{const Trueprop} $ (Const (@{const_name HOL.eq}, _) $ lhs $ _)) =
        let val head = head_of lhs in
          if member (op aconv) frees head then head else malformed_eqn_head lthy eqn
        end
      | head_fun_name_of_eqn eqn = malformed_eqn_lhs_rhs lthy eqn;

    val specss = map snd (AList.group (op aconv) (map (`(head_fun_name_of_eqn o snd)) specs));

    val eqn_bss = map (map (fst o fst)) specss;
    val eqn_attrsss = map (map (snd o fst)) specss;
    val eqnss = map (map snd) specss;

    val goalss = []; (* FIXME *)

    fun after_qed thmss lthy =
      let
        (* FIXME: temporary make_thm *)
        val simpss = map (map (Skip_Proof.make_thm (Proof_Context.theory_of lthy))) eqnss;

        fun mk_notes eqn_bs eqn_attrss fun_name qualify thms =
          let
            val anonymous_note =
              ((qualify (Binding.qualify true fun_name (Binding.name simpsN)), []), [(thms, [])]);
            val notes =
              @{map 3} (fn b => fn attrs => fn thm =>
                  ((Binding.qualify false fun_name b, code_attrs @ attrs), [([thm], [])]))
                eqn_bs eqn_attrss thms;
          in
            anonymous_note :: notes
          end;

        val notes = flat (@{map 5} mk_notes eqn_bss eqn_attrsss fun_names qualifys simpss);
      in
        ([] (* FIXME *), lthy |> Local_Theory.notes notes |> snd)
      end;
  in
    (goalss, after_qed, lthy)
  end;

fun gen_nonuniform_primrec prep_spec int =
  (fn (goalss, after_qed, lthy) => after_qed [] lthy) oooo
  gen_nonuniform_primrec_ursive prep_spec int false;

val nonuniform_primrec = gen_nonuniform_primrec Specification.check_multi_specs;
val nonuniform_primrec_cmd = gen_nonuniform_primrec Specification.read_multi_specs;

fun gen_nonuniform_primrecursive prep_spec int =
  (fn (goalss, after_qed, lthy) =>
     lthy
     |> Proof.theorem NONE (snd oo after_qed) goalss
     |> Proof.refine_singleton (Method.primitive_text (K I))) oooo
  gen_nonuniform_primrec_ursive prep_spec int true;

val nonuniform_primrecursive = gen_nonuniform_primrecursive Specification.check_multi_specs;
val nonuniform_primrecursive_cmd = gen_nonuniform_primrecursive Specification.read_multi_specs;

val nu_rec_option_parser = Parse.group (K "option")
  (Plugin_Name.parse_filter >> Plugins_NU_Rec_Option);

val _ = Outer_Syntax.local_theory @{command_keyword nonuniform_primrec}
  "define nonuniform primitive recursive functions"
  ((Scan.optional (@{keyword "("} |-- Parse.!!! (Parse.list1 nu_rec_option_parser)
      --| @{keyword ")"}) []) -- Parse_Spec.specification
   >> (fn (opts, (fixes, specs)) => snd o nonuniform_primrec_cmd true opts fixes specs));

val _ = Outer_Syntax.local_theory_to_proof @{command_keyword nonuniform_primrecursive}
  "define nonuniform primitive recursive functions"
  ((Scan.optional (@{keyword "("} |-- Parse.!!! (Parse.list1 nu_rec_option_parser)
      --| @{keyword ")"}) []) -- Parse_Spec.specification
   >> (fn (opts, (fixes, specs)) => nonuniform_primrecursive_cmd true opts fixes specs));

datatype nu_corec_option =
  Plugins_NU_Corec_Option of Proof.context -> Plugin_Name.filter;

fun gen_nonuniform_primcorec_ursive prep_spec int long_cmd opts raw_fixes raw_specs lthy =
  let
    val dups = duplicates (op =) (map (Binding.name_of o #1) raw_fixes);
    val _ = null dups orelse error ("Duplicate constant name " ^ quote (hd dups));

    val plugins = get_first (fn Plugins_NU_Corec_Option f => SOME (f lthy) | _ => NONE) (rev opts)
      |> the_default Plugin_Name.default_filter;

    val code_attrs =
      if plugins code_plugin then [Code.add_default_eqn_attrib Code.Equation] else [];

    val ((fixes, specs), _) = prep_spec raw_fixes raw_specs lthy;

    val (bs, mxs) = map_split (apfst fst) fixes;
    val frees = map (fst #>> Binding.name_of #> Free) fixes;
    val fun_names = map Binding.name_of bs;
    val qualifys = map (fold_rev (uncurry Binding.qualify o swap) o Binding.path_of) bs;

    fun head_fun_name_of_eqn
          (eqn as @{const Trueprop} $ (Const (@{const_name HOL.eq}, _) $ lhs $ _)) =
        let val head = head_of lhs in
          if member (op aconv) frees head then head else malformed_eqn_head lthy eqn
        end
      | head_fun_name_of_eqn eqn = malformed_eqn_lhs_rhs lthy eqn;

    val specss = map snd (AList.group (op aconv) (map (`(head_fun_name_of_eqn o snd)) specs));

    val eqn_bss = map (map (fst o fst)) specss;
    val eqn_attrsss = map (map (snd o fst)) specss;
    val eqnss = map (map snd) specss;

    val goalss = []; (* FIXME *)

    fun after_qed thmss lthy =
      let
        (* FIXME: temporary make_thm *)
        val ctr_thmss = map (map (Skip_Proof.make_thm (Proof_Context.theory_of lthy))) eqnss;

        fun mk_notes eqn_bs eqn_attrss fun_name qualify thms =
          let
            val anonymous_note =
              ((qualify (Binding.qualify true fun_name (Binding.name simpsN)), []), [(thms, [])]);
            val notes =
              @{map 3} (fn b => fn attrs => fn thm =>
                  ((Binding.qualify false fun_name b, code_attrs @ attrs), [([thm], [])]))
                eqn_bs eqn_attrss thms;
          in
            anonymous_note :: notes
          end;

        val notes = flat (@{map 5} mk_notes eqn_bss eqn_attrsss fun_names qualifys ctr_thmss);
      in
        ([] (* FIXME *), lthy |> Local_Theory.notes notes |> snd)
      end;
  in
    (goalss, after_qed, lthy)
  end;

fun gen_nonuniform_primcorec prep_spec int =
  (fn (goalss, after_qed, lthy) => after_qed [] lthy) oooo
  gen_nonuniform_primcorec_ursive prep_spec int false;

val nonuniform_primcorec = gen_nonuniform_primcorec Specification.check_multi_specs;
val nonuniform_primcorec_cmd = gen_nonuniform_primcorec Specification.read_multi_specs;

fun gen_nonuniform_primcorecursive prep_spec int =
  (fn (goalss, after_qed, lthy) =>
     lthy
     |> Proof.theorem NONE (snd oo after_qed) goalss
     |> Proof.refine_singleton (Method.primitive_text (K I))) oooo
  gen_nonuniform_primcorec_ursive prep_spec int true;

val nonuniform_primcorecursive = gen_nonuniform_primcorecursive Specification.check_multi_specs;
val nonuniform_primcorecursive_cmd = gen_nonuniform_primcorecursive Specification.read_multi_specs;

val nu_corec_option_parser = Parse.group (K "option")
  (Plugin_Name.parse_filter >> Plugins_NU_Corec_Option);

val _ = Outer_Syntax.local_theory @{command_keyword nonuniform_primcorec}
  "define nonuniform primitive corecursive functions"
  ((Scan.optional (@{keyword "("} |-- Parse.!!! (Parse.list1 nu_corec_option_parser)
      --| @{keyword ")"}) []) -- Parse_Spec.specification
   >> (fn (opts, (fixes, specs)) => snd o nonuniform_primcorec_cmd true opts fixes specs));

val _ = Outer_Syntax.local_theory_to_proof @{command_keyword nonuniform_primcorecursive}
  "define nonuniform primitive corecursive functions"
  ((Scan.optional (@{keyword "("} |-- Parse.!!! (Parse.list1 nu_corec_option_parser)
      --| @{keyword ")"}) []) -- Parse_Spec.specification
   >> (fn (opts, (fixes, specs)) => nonuniform_primcorecursive_cmd true opts fixes specs));

end;
